/* Copyright (C) 2003, Fredrik Kilander */
/* rotsim.c
 *
 * 11-jun-2003/FK Changed the way drAmp and drLvl are applied.
 *
 * 02-jun-2003/FK Temporarily rotsim_new.c as we work to remove dead
 *                stuff and make the drum floating and filtered. Maybe
 *                also put a cross-over filter in front of the horn.
 *
 * 23-mar-2003/FK Alas, I found it. It was neither quantization or skipping
 *                noise. It was a simple programming error, although it took
 *                some effort to find it. In function spot(), the weight of
 *                fr0 was calculated as the complement of fr1, as before.
 *                However, in difference to earlier versions, fr1 was now also
 *                multiplied by a scale factor (gain) which is not part
 *                of the complement. As a result, fr0 should not have been
 *                calculated using the value of fr1.
 *                fr1 = gain * (p - floor(p));
 *                fr0 = gain * (1.0 - fr1);           WRONG!
 *                fr0 = gain * (1.0 - p + floor(p));  Correct.
 *                
 *                Expanding the erroneous expression:
 *                fr0 = gain * (1.0 - gain * (p - floor(p)));
 *                fr0 = gain - gain * (p - floor(p));
 *                The cube appears to have beeen the primary cause of the
 *                edge harmonics.
 *
 * 21-mar-2003/FK Tried horn filter interpolation. Still noise. Locked the
 *                offset in spot() function and it disappeared. Conclusion:
 *                it IS skipping noise. Sh*t.
 *
 * 19-mar-2003/FK Tried post-filtering the output on the left channel to see
 *                if I could take the edges off the quant. noise. It did not
 *                work (although the filter did). The experimental code is
 *                conditioned on the symbol POST_LPF.
 *
 * 17-mar-2003/FK No it wasn't sample skipping, it was low-frequency 
 *                quantization noise that emanated from varying gain in the
 *                FIR weights. I nailed the gain floor to 1.0 and pre-
 *                normalized the filter definitions prior to mixing them.
 *                Essentially, if you are changing the gain of the
 *                signal, change on each, not just every nth sample.
 *
 * 15-mar-2003/FK Much problems with the effect; essentially I suspect
 *                'skipping noise' as the taps travel, but I am not really
 *                sure of the cause. I didn't have it before. Or at least,
 *                I didn't HEAR it before. Filter definitions changed from
 *                completely static to copy/mix in initTables().
 *
 * 10-mar-2003/FK Rewrite initTables() after hours of Excel excercises.
 *
 * 02-mar-2003/FK First successful test of new initTables() routine using
 *                66 point filters and simulated microphones.
 *
 * 25-feb-2003/FK Switching to float type, 17 pt bandpass filter for the horns.
 *                Too screechy.
 *
 * 2002-11-28/FK Introduced ALT_PANNING trying to be more similar to a
 *               famous product that I have heard.
 *
 * 2002-07-06/FK Increased horn filter to 11 points. 
 *
 * 2002-07-03/FK With the much more efficient fragment generator in place,
 *               I extended the horn filter to 18 slots and installed a
 *               6-point bandpass filter. Sounds better.
 *
 * 2002-02-23/FK Having played with the BPHORN on for two weeks, I
 *               undefined it, and it appears to sound better without it.
 *               There is a definite tradeoff between filter taps and
 *               Doppler effect. Herr Doppler wins, at least for now.
 *
 * 2002-02-09/FK Debugged and finished a first version of the drum filter.
 *               It does appear to drag some extra cycles, but not too much.
 *               The sound, however, worries me. I thought the unfiltered
 *               horns sounded rather good, but the extra bass provided
 *               by the low-passed drum gives a substantial bottom and
 *               emphasizes the need to actually filter the horns too.
 *
 * 2002-02-07/FK Rewrote the horn position logic to use fix-point increments
 *               instead of fragment counters. Cleaner code and better
 *               precision in the timing specifications was the nice result.
 *               Intend to do the same treatment for the drum, but this time
 *               I will apply the lowpass filter and conditionalize the code
 *               in case it turns out to be too slow.
 *
 * 2002-01-31/FK I needed 12 (!) slots in order to get some Doppler
 *               out of the filter. But then it started to sound real nice.
 *               The drum tremolo FX is being done, but not mixed into the
 *               output. Should make a setting for that.
 *               Rotor speeds could perhaps be made by having a bigger
 *               table and increment through it, rather than staying on
 *               the same for a number of samples. Slightly more overhead.
 *               Anyway a table of 1024 * 12 * 4 = 49152 bytes.
 *               A table of 2048 draws 98304 bytes.
 * 
 * 2002-01-24/FK Third version of the rotary speaker simulation.
 *               This one uses a delay line but has no moving taps. Instead,
 *               a FIR filter is used to tap the delay line. Motion of the
 *               tap point, and hence the Doppler FX, is obtained by shifting
 *               the distribution of the non-zero weights in the filter.
 *               Weight arrays are precomputed and looked up from a table.
 *               'Turning the horn' is then reduced to advancing the index
 *               in the weight array table. Also, it appears that the
 *               CPU gives good performance when multiplying and summing
 *               all those weights: likely because most of them are zero
 *               and but perhaps also because pipe-lines in the CPU can
 *               be fully efficient.
 */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include "main.h"
#include "config.h"

#include "rotsim.h"

#define MAX(A,B) (((A) < (B)) ? (B) : (A))
#define MIN(A,B) (((A) < (B)) ? (A) : (B))

/* Experimental alternative 20021128 */
#define ALT_PANNING

/* #define CLIP_DETECT */

#define HFIPOL			/* Horn filter linear interpolation */
#undef HFIPOL

#define LPDRUM			/* Low-pass drum */


#define BPHORN			/* Band-pass horn */


#ifdef BPHORN

#define HFDEFSIZE 17		/* Horn filter definition size */

#else

#define HFDEFSIZE 1		/* Horn filter definition size */

#endif /* BPHORN */


#define HFSPANSIZE  66
#define HFTABLESIZE 2048 /*2048*/	/* Nof horn filter table rows */
#define HFINCTBLMASK ((unsigned int) ((HFTABLESIZE << 16) - 1))

#ifdef HFIPOL
#define HFTABLESIZEMASK ((unsigned int) (HFTABLESIZE - 1))
#endif /* HFIPOL */

#ifdef LPDRUM

#define DRXOV			/* Drum crossover filter */

#define DFDEFSIZE      2	/* Drum filter definition size */
#define DFSPANSIZE     2	/* Drum filter fractional tap weights */
#define DFTABLESIZE 1024	/* Drum filter table size */
#define DFINCTBLMASK ((unsigned int) ((DFTABLESIZE << 16) - 1))

#endif /* LPDRUM */

#define DLYLNSPLS 512		/* Delay line samples */

/*
 *                      +y         pi/2
 *                   North wall      |
 *           +------------------------------------+
 *           |       <--- 74 cm -->               |
 *           |                       -            |
 *    West   | ^                /          \      | East
 *    wall   | |                                  | wall
 *           | 52 cm                              |
 *     pi -- | |              |     0,0 ==<  |    |--- 0 radians
 *           | V                                  |
 *     -x    |                                    |  +x
 *           |                  \          /      |
 *           |                       -            |
 *           |                                    |
 *           |                                    |
 *           +------------------------------------+
 *                   South wall      |
 *                      -y          3/2 pi
 *
 * Default model of the rotary speaker cabinet (horn section).
 * The horn rotates counter-clockwise. Delay times, attenuations and
 * filtering are calculated for shortest path (horn mouth to microphone),
 * and two or three wall reflections (horn, wall, microphone), depending
 * on the position of the microphone.
 * The speed of sound through air is calculated to 340 meters/second.
 * This yields a cm-to-samples factor of samples = cm * 220.5 / 340.0
 * at a samplerate of 22050 Hz.
 */

static double rcx =  0.0;	/* Rotor center coordinates */
static double rcy =  0.0;
static double hrd =  20.0;	/* Horn radius cm */

static double mLa = M_PI / 2.0;	/* Left mic angle */
static double mLr = 100.0;	/* Left mic cm from 0,0 */
static double mLx;		/* Left mic coordinates (set elsewhere) */
static double mLy;

static double mRa = M_PI /-2.0;	/* Right mic angle */
static double mRr = 100.0;	/* Right mic cm from 0.0 */
static double mRx;		/* Right mic coordinates (set elsewhere) */
static double mRy;

static double nwy =  26.0;	/* North wall y */
static double ewx =  28.0;	/* East wall x */
static double swy = -26.0;	/* South wall y */
static double wwx = -45.0; 	/* West wall x */

static double gFlrL = 0.5;	/* Gain floor left */
static double gFlrR = 0.5;
static int    gCoupled = 1;	/* Stereo compression (max wins) */

static double airSpeed = 340.0;	/* Meters per second */

#ifdef BPHORN
/* For gigging: very simple
#define FIR04
#define FIR14
*/
#define FIR00
#define FIR16

/*
 * The filter initializations are dummies. They are replaced in function
 * initTables ().
 */

float hornFilterDef[HFDEFSIZE] = {	/* Horn opening facing listener */
  1 /* ,0,0,0, 0,0,0,0, 0, 0,0,0,0, 0,0,0,0 */
};

float hornFilterDf2[HFDEFSIZE] = {	/* Horn opening away from listener */
  1 /* ,0,0,0, 0,0,0,0, 0, 0,0,0,0, 0,0,0,0 */
};

float filters[14][HFDEFSIZE] = {

  /* #0 Bandpass from fdes */
  {
   0.000260214,  /* [  0] */
  -4.78433e-05,  /* [  1] */
  -0.000786847,  /* [  2] */
  -0.00193316,  /* [  3] */
  -0.00257061,  /* [  4] */
  -0.00158094,  /* [  5] */
   0.00116342,  /* [  6] */
   0.0043009,  /* [  7] */
   0.00581348,  /* [  8] */
   0.00458298,  /* [  9] */
   0.00132555,  /* [ 10] */
  -0.00194022,  /* [ 11] */
  -0.00344027,  /* [ 12] */
  -0.00287155,  /* [ 13] */
  -0.00131745,  /* [ 14] */
  -8.61795e-05,  /* [ 15] */
   0.000361252   /* [ 16] */
  },

  /* #1 Bandpass from fdes */
  {
  -0.000440994,  /* [  0] */
  -0.000647606,  /* [  1] */
  -0.00101229,  /* [  2] */
  -0.0011199,  /* [  3] */
  -0.000498317,  /* [  4] */
   0.00101742,  /* [  5] */
   0.00308274,  /* [  6] */
   0.00494602,  /* [  7] */
   0.00581348,  /* [  8] */
   0.00527041,  /* [  9] */
   0.00351235,  /* [ 10] */
   0.00124863,  /* [ 11] */
  -0.000666903,  /* [ 12] */
  -0.00166352,  /* [ 13] */
  -0.00169492,  /* [ 14] */
  -0.00116652,  /* [ 15] */
  -0.000612226   /* [ 16] */
  },

  /* #2 Can't remember this one */
  {
  3.31208e-19 , /* [0] */
  -0.000392349, /* [1] */
  -0.00170577 , /* [2] */
  -0.00359296 , /* [3] */
  -0.00461278 , /* [4] */
  -0.00294459 , /* [5] */
  0.00169132  , /* [6] */
  0.0069338   , /* [7] */
  0.00924664  , /* [8] */
  0.0069338   , /* [9] */
  0.00169132  , /* [10] */
  -0.00294459 , /* [11] */
  -0.00461278 , /* [12] */
  -0.00359296 , /* [13] */
  -0.00170577 , /* [14] */
  -0.000392349, /* [15] */
  3.31208e-19   /* [16] */
  },

  /* #3 Another strange one */
  {
  -0.000173353, /* [0] */
  -0.00146133 , /* [1] */
  -0.0041557  , /* [2] */
  -0.00731959 , /* [3] */
  -0.00803062 , /* [4] */
  -0.00390001 , /* [5] */
  0.00434821  , /* [6] */
  0.0126357   , /* [7] */
  0.0161134   , /* [8] */
  0.0126357   , /* [9] */
  0.00434821  , /* [10] */
  -0.00390001 , /* [11] */
  -0.00803062 , /* [12] */
  -0.00731959 , /* [13] */
  -0.0041557  , /* [14] */
  -0.00146133 , /* [15] */
  -0.000173353  /* [16] */
  },

  /* #4 Unit tap */
  {
  1.0, 0.0, 0,
  0, 0, 0,
  0, 0, 0,
  0, 0, 0,
  0, 0, 0,
  0, 0
  },

  /* #5 Triangular window */
  {
   1, 2, 3, 4, 5, 6, 7, 8,
   9,
   8, 7, 6, 5, 4, 3, 2, 1
  },

  /* #6 Asymmetric bandpass definition */
  {
   0.1270179,			/* 0 */
   0.0871086,			/* 1 */
   0.0620930,			/* 2 */
   0.0011429,			/* 3 */
   -0.0728255,			/* 4 */
   -0.1261784,			/* 5 */
   -0.1319326,			/* 6 */
   -0.0846070,			/* 7 */
   -0.0040790,			/* 8 */
   0.0742398,			/* 9 */
   0.1174427,			/* 10 */
   0.1112052,			/* 11 */
   0.0651573,			/* 12 */
   0.0053796,			/* 13 */
   -0.0406771,			/* 14 */
   -0.0577677,			/* 15 */
   -0.0477476			/* 16 */
  },

  /* #7 Bandpass */
  {
  -0.000104732,  /* [  0] */
  -0.000196806,  /* [  1] */
  -0.000363396,  /* [  2] */
  -0.000479616,  /* [  3] */
  -0.000352733,  /* [  4] */
   0.000120921,  /* [  5] */
   0.000841208,  /* [  6] */
   0.0015215,  /* [  7] */
   0.00183838,  /* [  8] */
   0.00162129,  /* [  9] */
   0.00095844,  /* [ 10] */
   0.000148402,  /* [ 11] */
  -0.000472066,  /* [ 12] */
  -0.000712431,  /* [ 13] */
  -0.000608447,  /* [ 14] */
  -0.000354503,  /* [ 15] */
  -0.000145397   /* [ 16] */
  },

  /* #8 Weird creature */
  {
  -7.61452e-05,  /* [  0] */
  -5.60677e-05,  /* [  1] */
  -4.5468e-06,  /* [  2] */
   0.000154448,  /* [  3] */
   0.000454322,  /* [  4] */
   0.000866972,  /* [  5] */
   0.00130817,  /* [  6] */
   0.00166551,  /* [  7] */
   0.00183838,  /* [  8] */
   0.00177474,  /* [  9] */
   0.00149048,  /* [ 10] */
   0.001064,  /* [ 11] */
   0.000608024,  /* [ 12] */
   0.00022942,  /* [ 13] */
  -7.61285e-06,  /* [ 14] */
  -0.000100994,  /* [ 15] */
  -0.000105711   /* [ 16] */
  },

  /* #9 Probably weak bandpass */
  {
  3.19317e-19 , /* [0] */
  -0.000257316, /* [1] */
  -0.000835697, /* [2] */
  -0.00139371 , /* [3] */
  -0.00145145 , /* [4] */
  -0.000657322, /* [5] */
  0.000832002 , /* [6] */
  0.0023041   , /* [7] */
  0.0029188   , /* [8] */
  0.0023041   , /* [9] */
  0.000832002 , /* [10] */
  -0.000657322, /* [11] */
  -0.00145145 , /* [12] */
  -0.00139371 , /* [13] */
  -0.000835697, /* [14] */
  -0.000257316, /* [15] */
  3.19317e-19   /* [16] */
  },

  /* #10 Average filter */
  {
  0.1, 0.1, 0.1,
  0.1, 0.1, 0.1,
  0.1, 0.1, 0.1,
  0.1, 0.1, 0.1,
  0.1, 0.1, 0.1,
  0.1, 0.1
  },

  /* #11 Another unit tap */
  {
  1.0, 0.0, 0,
  0, 0, 0,
  0, 0, 0,
  0, 0, 0,
  0, 0, 0,
  0, 0
  },

  /* #12 Muted tap */
  {
  0.0, 0.0, 0,
  0, 0, 0,
  0, 0, 0,
  0, 0, 0,
  0, 0, 0,
  0, 0
  },

  /* #13 1800 Hz bandpass from fdes with more digits */
  {
  -0.000086384595849389, /* [0] */
  -0.000210152725922792, /* [1] */
  -0.000456282814522991, /* [2] */
  -0.000695933549366180, /* [3] */
  -0.000680880734970253, /* [4] */
  -0.000255271712241711, /* [5] */
  0.000477820340204896, /* [6] */
  0.001175837946738534, /* [7] */
  0.001462495691859744, /* [8] */
  0.001175837946738562, /* [9] */
  0.000477820340204896, /* [10] */
  -0.000255271712241725, /* [11] */
  -0.000680880734970253, /* [12] */
  -0.000695933549366179, /* [13] */
  -0.000456282814522991, /* [14] */
  -0.000210152725922792, /* [15] */
  -0.000086384595849389  /* [16] */
  }
};

static double hfDefGain = 0.01;
static double hfDf2Gain = 0.01;
static double hornGain  = 0.2; /* 0.2; */

#else

static float hornFilterDef [HFDEFSIZE] = {1.0};
static float hornFilterDf2 [HFDEFSIZE] = {1.0};

#endif /* BPHORN */


static float hornFilterTb1 [HFTABLESIZE][HFSPANSIZE];
static float hornFilterTb2 [HFTABLESIZE][HFSPANSIZE];

static double hornRPMSlow      =  40.0;
static double hornRPMFast      = 400.0;
static double hornRevUpSeconds =   1.0;
static double hornRevDnSeconds =   0.7;

static unsigned int hornPos = 0;

static int hornPosIncrement = 0;
static int hornPosIncSlow = 0;
static int hornPosIncFast = 0;

static int hornAcceleration = 0;
static int hornDeceleration = 0;

static int hornSpeedChange = 0;

#ifdef LPDRUM

static double drumRPMSlow =  30.0;
static double drumRPMFast = 330.0;

static double drumRevUpSeconds = 3.2;
static double drumRevDnSeconds = 3.9;

static double drLvl = 0.25;	/* Drum level */
static double drAmp = 0.25;	/* Drum amplitude modulation */

static float drumFilterDef[DFDEFSIZE] = {0.5, 0.5};
static float drumFilter [DFTABLESIZE][DFSPANSIZE];

static unsigned int drumPos = 0;
static int drumPosIncrement = 0;
static int drumPosIncSlow = 0;
static int drumPosIncFast = 0;
static int drumAcceleration = 0;
static int drumDeceleration = 0;
static int drumSpeedChange = 0;

#ifdef DRXOV

static float drxA[6] = {
  0.00690755623266794960279924, /* a0 in section 1 */
  1.70356998430480111039742042, /* a1 in section 1 */
  -0.73120020923547290880861738, /* a2 in section 1 */
  0.01136479022253778121964718,  /* a0 in section 2 */
  1.83415628681387565812599405,  /* a1 in section 2 */
  -0.87961544770402677606568886  /* a2 in section 2 */
};

static float drxZ[4];

#endif /* DRXOV */

#endif /* LPDRUM */

static int speedIsFast = FALSE;

/* Delay line for the horn */

static float delay [DLYLNSPLS + 1];	/* Delay line */
static float * delayEnd;	/* End of delay line */
static float * dp;		/* Input position */
static float * tap1;		/* Position of last weight in 1st tap */
static float * tap2;		/* Position of last weight in 2nd tap */
static float * tapGuard;	/* Taps must wrap from this position */

/* Delay line for the drum */

#ifdef LPDRUM
static float   drDelay [DLYLNSPLS + 2];
static float * drDelayEnd;
static float * dtap;		/* Drum tap */
static float * dtapGuard;	/* Drum tap guard */
#endif /* LPDRUM */

static double bwAmp = 0.775;	/* Tap2: 'Back wall' attenuation. */
static double lwAmp = 0.775;	/* Tap2: 'Left wall' attenuation */
static double rwAmp = 0.0;//-0.775;	/* Tap2: 'Left wall' attenuation */

/* ================================================================ */

void setHornRPMSlow (double rpm) { hornRPMSlow = rpm;}
void setHornRPMFast (double rpm) { hornRPMFast = rpm;}
void setHornRevUpSec (double sec) { hornRevUpSeconds = sec;}
void setHornRevDnSec (double sec) { hornRevDnSeconds = sec;}

void setDrumRPMSlow (double rpm) {
#ifdef LPDRUM
  drumRPMSlow = rpm;
#endif
}

void setDrumRPMFast (double rpm) {
#ifdef LPDRUM
  drumRPMFast = rpm;
#endif
}

void setDrumRevUpSec (double seconds) {
#ifdef LPDRUM
  drumRevUpSeconds = seconds;
#endif
}

void setDrumRevDnSec (double seconds) {
#ifdef LPDRUM
  drumRevDnSeconds = seconds;
#endif
}

static void setDrumLevel (double level) {
#ifdef LPDRUM
  drLvl = level;
#endif
}

static void setDrumAmplMod (double mod) {
#ifdef LPDRUM
  if ((0.0 <= mod) && (mod <= 1.0)) {
    drAmp = mod;
  }
#endif
}

static void setBackWallReflection (double level) {
  bwAmp = level;
}

static void setLeftWallReflection (double level) {
  lwAmp = level;
}

static void setRightWallReflection (double level) {
  rwAmp = level;
}

/*
 *
 */
int rotConfig (ConfigContext * cfg) {
  if (strcasecmp (cfg->name, "rotsim.horn.slowrpm") == 0) {
    double r;
    if (sscanf (cfg->value, "%lf", &r) == 1) {
      setHornRPMSlow (r);
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "rotsim.horn.fastrpm") == 0) {
    double r;
    if (sscanf (cfg->value, "%lf", &r) == 1) {
      setHornRPMFast (r);
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "rotsim.horn.revupsec") == 0) {
    double s;
    if (sscanf (cfg->value, "%lf", &s) == 1) {
      setHornRevUpSec (s);
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "rotsim.horn.revdnsec") == 0) {
    double s;
    if (sscanf (cfg->value, "%lf", &s) == 1) {
      setHornRevDnSec (s);
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "rotsim.drum.slowrpm") == 0) {
    double r;
    if (sscanf (cfg->value, "%lf", &r) == 1) {
      setDrumRPMSlow (r);
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "rotsim.drum.fastrpm") == 0) {
    double r;
    if (sscanf (cfg->value, "%lf", &r) == 1) {
      setDrumRPMFast (r);
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "rotsim.drum.revupsec") == 0) {
    double s;
    if (sscanf (cfg->value, "%lf", &s) == 1) {
      setDrumRevUpSec (s);
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "rotsim.drum.revdnsec") == 0) {
    double s;
    if (sscanf (cfg->value, "%lf", &s) == 1) {
      setDrumRevDnSec (s);
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "rotsim.drum.x-amplmod") == 0) {
    double a;
    if (sscanf (cfg->value, "%lf", &a) == 1) {
      setDrumAmplMod (a);
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "rotsim.drum.x-level") == 0) {
    double a;
    if (sscanf (cfg->value, "%lf", &a) == 1) {
      setDrumLevel (a);
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "rotsim.x-backwallreflection") == 0) {
    double a;
    if (sscanf (cfg->value, "%lf", &a) == 1) {
      setBackWallReflection (a);
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "rotsim.x-leftwallreflection") == 0) {
    double a;
    if (sscanf (cfg->value, "%lf", &a) == 1) {
      setLeftWallReflection (a);
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "rotsim.x-rightwallreflection") == 0) {
    double a;
    if (sscanf (cfg->value, "%lf", &a) == 1) {
      setRightWallReflection (a);
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "rotsim.mic.distance") == 0) {
    double a;
    if (sscanf (cfg->value, "%lf", &a) == 1) {
      mLr = mRr = a;			/* should be set-method */
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "rotsim.mic.left.angle") == 0) {
    double a;
    if (sscanf (cfg->value, "%lf", &a) == 1) {
      mLa = a;			/* should be set-method */
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "rotsim.mic.left.distance") == 0) {
    double a;
    if (sscanf (cfg->value, "%lf", &a) == 1) {
      mLr = a;			/* should be set-method */
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "rotsim.mic.right.angle") == 0) {
    double a;
    if (sscanf (cfg->value, "%lf", &a) == 1) {
      mRa = a;			/* should be set-method */
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "rotsim.mic.right.distance") == 0) {
    double a;
    if (sscanf (cfg->value, "%lf", &a) == 1) {
      mRr = a;			/* should be set-method */
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "rotsim.mic.left.min-gain") == 0) {
    double a;
    if (sscanf (cfg->value, "%lf", &a) == 1) {
      gFlrL = a;		/* should be set-method */
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "rotsim.mic.right.min-gain") == 0) {
    double a;
    if (sscanf (cfg->value, "%lf", &a) == 1) {
      gFlrR = a;		/* should be set-method */
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "rotsim.mic.min-gain") == 0) {
    double a;
    if (sscanf (cfg->value, "%lf", &a) == 1) {
      gFlrL = gFlrR = a;	/* should be set-method */
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "rotsim.horn.radius") == 0) {
    double a;
    if (sscanf (cfg->value, "%lf", &a) == 1) {
      hrd = a;	/* should be set-method */
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
}

/*
 * Convert from cm to samples; the unit by which sound has to travel.
 */
static double cm2spl (double cm, double S) {
  return (cm * S * 0.01) / airSpeed;
}

/*
 * Compute the distance from the microphone to the mouth of the horn
 * at a given rotation angle of the horn.
 * Not used at the moment.
 * @param phi   The rotation angle of the horn. The horn is facing the mic
 *              at an angle of 0.0 .
 * @param r     The radius of the horn.
 * @param dc    The distance from the mic to the rotation center of the horn.
 *
 * @return      The distance from the mic to the horn opening.
 */

static double dst2hm (double phi, double r, double dc) {
  double a = dc + (r * (1.0 - cos (phi)));
  double b = r * sin (phi);
  return sqrt ((a * a) + (b * b));
}

/*
 * Compute standard Euclidian distance between two points in a
 * coordinate system.
 */

static double distance (double x0, double y0, double x1, double y1) {
  return sqrt (((x1 - x0) * (x1 - x0)) + ((y1 - y0) * (y1 - y0)));
}

static double distance3 (double x0,
			 double y0,
			 double x1,
			 double y1,
			 double x2,
			 double y2) {
  return distance (x0, y0, x1, y1) + distance (x1, y1, x2, y2);
}

/*
 * Computes the horn eq selection angle.
 * @param ha  Current horn rotation angle
 * @param hx  Horn mouth x
 * @param hy  Horn mouth y
 * @param px  Point x
 * @param py  Point y
 * @return The angle relative to the horn in the range 0--2pi
 */

static double angle (double ha, double hx, double hy, double px, double py) {
  double r = atan2 (py - hy, px - hx) - ha;
  while (r < 0.0) {
    r += 2.0 * M_PI;
  }
  return r;
}

/*
 * Return the relative mix between front and back sound.
 * @return 1.0 : facing horn mouth, 0.0 : facing counterweight.
 */

static double relMix (double ha, double hx, double hy, double px, double py) {
  return 0.5 + (0.5 * cos (angle (ha, hx, hy, px, py)));
}

/*
 * @param radius  Rotor radius
 * @param direct  Direct sound distance
 * @param bounce  Bounced sound distance
 */

static double bounceGain (double radius, double direct, double bounce) {

  double bg;

  if (fabs (bounce - direct) < 1.0e-9) {
    assert (0.0 < direct);
    bg = (radius * radius) / (direct * direct);
  }
  else {
    assert (0.0 < bounce);
    bg = (radius * radius) / (bounce * bounce);
  }

#if 0
  printf ("bg(r=%g, d=%g, b=%g) = %g\n", radius, direct, bounce, bg);
#endif

  return bg;
}

/*
 * Inverted gain: drops with inverted squared distance. Actually, higher
 * frequencies (wherever _they_ start) drops with the cube because of
 * 'dispersion', presumable in the air. So a more correct model would
 * probably have to return a filter. Something we actually can do if
 * we like.
 */

static double invGain (double dist) {
  return -1.0 / (1.0 - (dist * dist));
}

/*
 * Compensated (or compressed) gain.
 */

static double compGain (double gFloor,
			double gRange,
			double gMin,
			double distance) {
  return gFloor + ((1.0 - gFloor) * ((invGain (distance) - gMin) / gRange));
}

/*
 * This function computes and returns the balance between the front and
 * aft filter coefficient.
 * @param rmix The output for the relmix() function.
 */

float filterBalance (double rmix, int j) {
#if 1
  return (rmix * hornFilterDef[j]) + ((1.0 - rmix) * hornFilterDf2[j]);
#else
  return (rmix * hornFilterDef[j]) - ((1.0 - rmix) * hornFilterDf2[j]);
#endif
}

/*
 * Helper function for initTables ().
 */
#define COERCE_P
static int spot (double p,	/* Preferred delay position */
		 int   fdp,	/* Position in filter definition */
		 int * t0,
		 int * t1,
		 double * fr0,
		 double * fr1,
		 double gain) 
{
  int t = (int) floor (p);	/* Index to first element in FIR */
  assert (t0 != NULL);
  assert (t1 != NULL);
  assert (fr0 != NULL);
  assert (fr1 != NULL);

#if 0				/* DEBUG */
  t = 2;			/* Lock position */
  p = (double) t;
#endif

  if (HFSPANSIZE < (t + HFDEFSIZE)) {
    /* We can't extend further to the right (into the past) */
    t = HFSPANSIZE - HFDEFSIZE;
#ifdef COERCE_P
    p = (double) t;
#endif
  }
  else if (t < 0) {
    /* We can't extend further to the left (into the future) */
    t = 0;
#ifdef COERCE_P
    p = (double) t;
#endif
  }

  *t0  = t + fdp;
  *t1  = t + fdp + 1;

  *fr1 = gain * (p - floor (p)); /* Fraction of FIR weight [t1] */
  *fr0 = gain * (1.0 - p + floor (p)); /* Fraction of FIR weight [t0] */

  /* Look out for special case at end of filter definition. */

  if (HFSPANSIZE <= (*t1)) {
    *t1  = *t0;
    *fr1 = 0.0;
  }

  return 0;
}

/*
 * This function initializes the FIR filter tables for the delay lines.
 */
#define DD 0
#define NW 1
#define EW 2
#define SW 3
#define WW 4

static void initTables () {

  static double fullCircle = 2.0 * M_PI;

  int acL[5];			/* Left mic: active paths */
  int acR[5];			/* Right mic: active paths */

  double bpxL[5];		/* Bounce point coordinates for computing */
  double bpyL[5];		/* the angle between a path and the mouth */
  double bpxR[5];		/* of the horn. */
  double bpyR[5];

  int i;
  int j;
  double a;			/* Instant horn rotation in radians */

  double hx;			/* Instant horn coordinates */
  double hy;

  double pLx;			/* Left mic: refl. x in north & south walls */
  double pLy;			/* Left mic: refl. y in east and west walls */

  double pRx;			/* Right mic:refl. x in north & south walls */
  double pRy;			/* Right mic:refl. y in east and west walls */

  double dnL[5];		/* Left mic. distances: Direct, N, E, S, W */
  double dnR[5];

  double gMinL;			/* Left mic: Minimum gain from distance */
  double gMaxL;			/* Left mic: Maximum gain from distance */
  double gRngL;			/* Left mic: Dynamic range of gain */
  double gnL[5];		/* Left mic: gain from 5 sources */

  double gMinR;
  double gMaxR;
  double gRngR;
  double gnR[5];

  double dLr;			/* Left mic: distance to horn radius */
  double dRr;			/* Right mic: distance to horn radius */
  double dMin;			/* The smallest of the two. */

  double sum;

  double dTabMax[2];
  double dTabMin[2];
  double normGain;

  /* Initialize filter definitions */

  for (i = 0; i < HFDEFSIZE; i++) {
/*     hornFilterDef[i] = filters[ 6][i] + (0.0 * filters[10][i]); */
/*     hornFilterDf2[i] = filters[ 8][i] + (0.0 * filters[10][i]); */
    hornFilterDef[i] = filters[13][i] * 60.0;
    hornFilterDf2[i] = filters[13][i] * 40.0; /* Reflection */
  }

#if 0
  /*
   * 17-mar-2003/FK This stays in.
   * This is an attempt to pre-normalize the filter definitions in order
   * to prevent quantization noise as we crossfade from one filter to
   * another. The composite filters will be normalized again below.
   */
  {
    double sum0 = 0.0;
    double sum1 = 0.0;
    for (i = 0; i < HFDEFSIZE; i++) {
      sum0 += hornFilterDef[i];
      sum1 += hornFilterDf2[i];
    }
    for (i = 0; i < HFDEFSIZE; i++) {
      hornFilterDef[i] /= sum0;
      hornFilterDf2[i] /= sum1;
    }
    printf ("\nHORN FLTR DEF TBL 1 : SUM = %g\n", sum0);
    printf ("HORN FLTR DEF TBL 2 : SUM = %g\n", sum1);
  }
#endif

  /* Compute the microphone coordinates */

  mLx = mLr * cos (mLa);
  mLy = mLr * sin (mLa);
  mRx = mRr * cos (mRa);
  mRy = mRr * sin (mRa);

  /* Determine active paths */

  acL[DD] = TRUE;		/* Always */
  acL[NW] = mLy < nwy;
  acL[EW] = mLx < ewx;
  acL[SW] = swy < mLy;
  acL[WW] = wwx < mLx;

  acR[DD] = TRUE;
  acR[NW] = mRy < nwy;
  acR[EW] = mRx < ewx;
  acR[SW] = swy < mRy;
  acR[WW] = wwx < mRx;

  /* Compute distances under the current geometry */

  for (i = 0; i < HFTABLESIZE; i++) {

    a = (fullCircle * (double) i) / (double) HFTABLESIZE;

    hx = rcx + (hrd * cos (a));	/* Horn coordinates */
    hy = rcy + (hrd * sin (a));

    pLx = (hx + mLx) / 2.0;	/* Reflection points */
    pLy = (hy + mLy) / 2.0;
    pRx = (hx + mRx) / 2.0;
    pRy = (hy + mRy) / 2.0;

    dnL[DD] = distance (hx, hy, mLx, mLy); /* Direct distance */
    dnL[NW] = distance3 (hx, hy, pLx, nwy, mLx, mLy); /* North reflection */
    dnL[EW] = distance3 (hx, hy, ewx, pLy, mLx, mLy); /* East reflection */
    dnL[SW] = distance3 (hx, hy, pLx, swy, mLx, mLy); /* South reflection */
    dnL[WW] = distance3 (hx, hy, wwx, pLy, mLx, mLy); /* West reflection */

    dnR[DD] = distance (hx, hy, mRx, mRy);
    dnR[NW] = distance3 (hx, hy, pRx, nwy, mRx, mRy);
    dnR[EW] = distance3 (hx, hy, ewx, pRy, mRx, mRy);
    dnR[SW] = distance3 (hx, hy, pRx, swy, mRx, mRy);
    dnR[WW] = distance3 (hx, hy, wwx, pRy, mRx, mRy);

    if (i == 0) {		/* Initialize once */
      gMinL = gMaxL = invGain (dnL[DD]);
      gMinR = gMaxR = invGain (dnR[DD]);
    }

    for (j = DD; j <= WW; j++) { /* Find min and max gains for two mics */
      double g;

      if (acL[j]) {
	g = invGain (dnL[j]);
	if (g < gMinL) gMinL = g;
	if (gMaxL < g) gMaxL = g;
      }

      if (acR[j]) {
	g = invGain (dnR[j]);
	if (g < gMinR) gMinR = g;
	if (gMaxR < g) gMaxR = g;
      }
    }
    
  } /* for a full circle of the horn */

  /* Compute normalization values for the gain changes determined by
     distance between the mic and horn mouth (via reflection). These
     gain changes are dramatic close to the horn and very small when
     farther off. In order to preserve audibility we let the parameters
     gFlrL and gFlrR establish a gain floor (minimum gain) and then
     compress the dynamic range provided by the rotation in the space
     between the floor and unity (1.0). */

  gRngL = gMaxL - gMinL;	/* Ranges will typically be very small */
  gRngR = gMaxR - gMinR;

  if (gCoupled) {		/* If stereo compression let the mic */
    if (gRngL < gRngR) {	/* with the greatest range determine */
      gRngL = gRngR;		/* the normal range for both mics. */
    }
    else {
      gRngR = gRngL;
    }
  }

  /* All distances include the shortest distance which is the distance */
  /* between a mic and the horn radius. That distance will never play out */
  /* in the FIRs so we want to do away with it. */

  /* Distance from left mic to rotor radius */

  dLr = distance (rcx, rcy, mLx, mLy) - hrd;

  /* Distance from right mic to rotor radius */

  dRr = distance (rcx, rcy, mRx, mRy) - hrd;

  /* The smallest value of the two */

  dMin = (dLr < dRr) ? dLr : dRr;

  /* Adjust distances for FIR indexing */

  dLr = dMin;
  dRr = dMin;

  /*
   * Width is the number of samples available for motion in the delay
   * line. With a wider FIR filter tap, the width of the motion is
   * smaller:      [. . . . . <-- {FIR taps} --> . . . . . .]
   * Thus width is the span of the delay line minus the span of the taps.
   * It is important to remember that the tap itself is the same length
   * as the delay line, but the tap definition is shorter. The tap is
   * 'moved' by replacing zeros in the coefficent lookup table. For example,
   * assume a delay line of 5 samples:          [  s0,  s1,  s2,  s3,  s4 ]
   * and the left-most FIR coefficient array:   [ 1.0, 0.0, 0.0, 0.0, 0.0 ]
   * To advance the tap half a sample right:    [ 0.5, 0.5, 0.0, 0.0, 0.0 ]
   * This example has a width of 4 because the tap only covers 1 sample,
   * even though it may be distributed across several samples.
   * Then to further complicate the situation, several taps have their
   * coefficients mixed into the same lookup table. In the spirit of the
   * above example, two taps, one non-inverting and one inverting move towards
   * each other at half the sampling rate:      [ 1.0, 0.0, 0.0, 0.0,-1.0 ]
   *                                            [ 0.5, 0.5, 0.0,-0.5,-0.5 ]
   *                                            [ 0.0, 1.0, 0.0,-1.0, 0.0 ]
   *                                            [ 0.0, 0.5, 0.0,-0.5, 0.0 ]
   */

#if 0				/* DEBUG */
  for (j = DD; j <= WW; j++) {
    acL[j] = acR[j] = FALSE;
  }
  j = DD;
  acL[j] = TRUE;
  acR[j] = FALSE;
#endif

  for (i = 0; i < HFTABLESIZE; i++) {
    /* a is the fraction of a full circle (in radians) mapped by i. */
    a = (fullCircle * (double) i) / (double) HFTABLESIZE;

    hx = rcx + (hrd * cos (a));	/* Horn mouth x coordinate */
    hy = rcy + (hrd * sin (a));	/* Horn mouth y coordinate */

    pLx = (hx + mLx) / 2.0; /* Bounce point x north south walls */
    pLy = (hy + mLy) / 2.0; /* Bounce point y west east walls */
    pRx = (hx + mRx) / 2.0;
    pRy = (hy + mRy) / 2.0;

    bpxL[DD] = mLx; bpyL[DD] = mLy;
    bpxL[NW] = pLx; bpyL[NW] = nwy;
    bpxL[EW] = ewx; bpyL[EW] = pLy;
    bpxL[SW] = pLx; bpyL[SW] = swy;
    bpxL[WW] = wwx; bpyL[WW] = pLy;

    bpxR[DD] = mRx; bpyR[DD] = mRy;
    bpxR[NW] = pRx; bpyR[NW] = nwy;
    bpxR[EW] = ewx; bpyR[EW] = pRy;
    bpxR[SW] = pRx; bpyR[SW] = swy;
    bpxR[WW] = wwx; bpyR[WW] = pRy;

    dnL[DD] = distance (hx, hy, mLx, mLy); /* Direct distance */
    dnL[NW] = distance3 (hx, hy, pLx, nwy, mLx, mLy); /* North reflection */
    dnL[EW] = distance3 (hx, hy, ewx, pLy, mLx, mLy); /* East reflection */
    dnL[SW] = distance3 (hx, hy, pLx, swy, mLx, mLy); /* South reflection */
    dnL[WW] = distance3 (hx, hy, wwx, pLy, mLx, mLy); /* West reflection */

    dnR[DD] = distance (hx, hy, mRx, mRy);
    dnR[NW] = distance3 (hx, hy, pRx, nwy, mRx, mRy);
    dnR[EW] = distance3 (hx, hy, ewx, pRy, mRx, mRy);
    dnR[SW] = distance3 (hx, hy, pRx, swy, mRx, mRy);
    dnR[WW] = distance3 (hx, hy, wwx, pRy, mRx, mRy);

    /* Compute compressed gains */

    for (j = DD; j <= WW; j++) {
      gnL[j] = compGain (gFlrL, gRngL, gMinL, dnL[j]);
      gnR[j] = compGain (gFlrR, gRngR, gMinR, dnR[j]);
    }


    for (j = 0; j < HFDEFSIZE; j++) {
      int k;
      double p;
      int    t0;		/* Target element in filter */
      int    t1;		/* Next element (target +1) */
      double fr0;		/* Fraction of target cell covered */
      double fr1;		/* Complement to frac */

      for (k = DD; k <= WW; k++) {

	if (acL[k]) {
	  p = cm2spl (dnL[k] - dLr, SampleRateD);
	  spot (p, j, &t0, &t1, &fr0, &fr1, gnL[k]);

#if 0
	  sum = filterBalance (compGain (0.0, gRngL, gMinL, dnL[k]), j);
#else
	  sum = filterBalance (relMix (a, hx, hy, bpxL[k], bpyL[k]), j);
#endif
	  hornFilterTb1[i][t0] += fr0 * sum;
	  hornFilterTb1[i][t1] += fr1 * sum;
	}

	if (acR[k]) {
	  p = cm2spl (dnR[k] - dRr, SampleRateD);
	  spot (p, j, &t0, &t1, &fr0, &fr1, gnR[k]);

#if 0
	  sum = filterBalance (compGain (0.0, gRngR, gMinR, dnR[k]), j);
#else
	  sum = filterBalance (relMix (a, hx, hy, bpxR[k], bpyR[k]), j);
#endif
	  hornFilterTb2[i][t0] += fr0 * sum;
	  hornFilterTb2[i][t1] += fr1 * sum;
	}
      }

    } /* for j */
  } /* for i */

  /* Measure and normalize the tables */

  dTabMin[0] = 0.0;
  dTabMin[1] = 0.0;

  for (j = 0; i < HFSPANSIZE; j++) {
    dTabMin[0] += hornFilterTb1[0][j];
    dTabMin[1] += hornFilterTb2[0][j];
  }

  dTabMax[0] = dTabMin[0];
  dTabMax[1] = dTabMin[1];

  for (i = 1; i < HFTABLESIZE; i++) {
    double sum1 = 0.0;
    double sum2 = 0.0;
    for (j = 0; j < HFSPANSIZE; j++) {
      /* Normalize by real or absolute value? */
#if 1
      sum1 += hornFilterTb1[i][j];
      sum2 += hornFilterTb2[i][j];
#else
      sum1 += fabs (hornFilterTb1[i][j]);
      sum2 += fabs (hornFilterTb2[i][j]);
#endif
    }
    if (sum1 < dTabMin[0]) dTabMin[0] = sum1;
    if (dTabMax[0] < sum1) dTabMax[0] = sum1;
    if (sum2 < dTabMin[1]) dTabMin[1] = sum2;
    if (dTabMax[1] < sum2) dTabMax[1] = sum2;
  }

  sum = MAX (dTabMax[0], dTabMax[1]);

  if (sum < 1.0e-16) {
    sum = 1.0;
  }

  normGain = hornGain / sum;

#if 0
  printf ("\nHORN TABLE 1 MIN=%g MAX=%g\n", dTabMin[0], dTabMax[0]);
  printf ("HORN TABLE 2 MIN=%g MAX=%g\n", dTabMin[1], dTabMax[1]);
  printf ("NORMALIZATION GAIN: %g = %g/%g\n", normGain, hornGain, sum);
  printf ("LEFT  MIC: X=%g Y=%g\n", mLx, mLy);
  printf ("RIGHT MIC: X=%g Y=%g\n", mRx, mRy);
#endif

  for (i = 0; i < HFTABLESIZE; i++) {
    for (j = 0; j < HFSPANSIZE; j++) {
#if 0
      hornFilterTb1[i][j] *= normGain;
      hornFilterTb2[i][j] *= normGain;
#endif
#if 0
      hornFilterTb1[i][j] *= hornGain / dTabMax[0];
      hornFilterTb2[i][j] *= hornGain / dTabMax[1];
#endif
    }
  }

#if 0
 {
   char * fname = "ROTFIR.txt";
   FILE * fp = fopen (fname, "w");
   if (fp == NULL) {
     perror (fname);
     assert (0);
   }

   fprintf (fp, "Tb1 (left)\n");

   for (i = 0; i < HFTABLESIZE; i++) {
     fprintf (fp, "%4d : ", i);
     for (j = 0; j < HFSPANSIZE; j++) {
       float fv = hornFilterTb1[i][j];
       char ch = '0';
       if (fv < 0.0) {
	 ch = '-';
       }
       else if (0.0 < fv) {
	 ch = '+';
       }
       fprintf (fp, "%c", ch);
     }
     fprintf (fp, "\n");
   }

   fprintf (fp, "Tb2 (right)\n");

   for (i = 0; i < HFTABLESIZE; i++) {
     fprintf (fp, "%4d : ", i);
     for (j = 0; j < HFSPANSIZE; j++) {
       float fv = hornFilterTb2[i][j];
       char ch = '0';
       if (fv < 0.0) {
	 ch = '-';
       }
       else if (0.0 < fv) {
	 ch = '+';
       }
       fprintf (fp, "%c", ch);
     }
     fprintf (fp, "\n");
   }

   for (i = 0; i < HFTABLESIZE; i++) {
     fprintf (fp, "%4d : ", i);
     for (j = 0; j < HFSPANSIZE; j++) {
       float w = hornFilterTb1[i][j];
       if (w != 0.0) {
	 fprintf (fp, "w[%2d]={%g}", j, w);
       }
     }
     fprintf (fp, "\n");
   }

#if 0
   fprintf (fp, "%10s%10s%10s%10s%10s\n", "MIN", "MAX", "SUM", "AVG", "UNIT");

   for (i = 0; i < HFTABLESIZE; i++) {
     double sum = 0.0;
     double min;
     double max;
     min = max = hornFilterTb1[i][0];
     fprintf (fp, "%4d : ", i);
     for (j = 1; j < HFSPANSIZE; j++) {
       float fv = hornFilterTb1[i][j];
       if (fv < min) min = fv;
       if (max < fv) max = fv;
       sum += fv;
     }
     fprintf (fp, "% 10.7f% 10.7f% 10.7f% 10.7f\n", min, max, sum, 
	      sum / (double) HFSPANSIZE);
   }
#endif

   fclose (fp);
 }
#endif

#ifdef LPDRUM

  /* This is a simple amplitude modulation of the drumfilter coefficients */

  for (i = 0; i < DFTABLESIZE; i++) {
    double u = sin ((fullCircle * (double) i) / (double) DFTABLESIZE);
    double g = drLvl * ((1.0 - drAmp) + ((drAmp / 2.0) * (1.0 + u)));
    for (j = 0; j < DFSPANSIZE; j++) {
      drumFilter[i][j] = (float) (g * (double) drumFilterDef[j]);
    }
  }

#endif /* LPDRUM */

}


/*
 * Returns the increment needed to traverse a certain table the
 * requested number of revolutions per minute.
 * @param tblSz  The number of slots in the table.
 * @param rpm    Revolutions per minute.
 */
static double getTableIncrement (size_t tblSz, double rpm) {
  return (rpm * (double) tblSz) / (SampleRateD * 60.0);
}

/*
 *
 */
void initRotSim () {
  static double sz = 128.0;	/* Fragment size in samples */
  double a;
  initTables ();

  hornSpeedChange = 0;

  hornPosIncSlow = (int)
    (getTableIncrement ((size_t) HFTABLESIZE, hornRPMSlow) * 65536.0);
  hornPosIncFast = (int)
    (getTableIncrement ((size_t) HFTABLESIZE, hornRPMFast) * 65536.0);

  hornPosIncrement = hornPosIncSlow;
  speedIsFast = FALSE;

  /* We change the speed between fragments, so the time is dependent
     on the fragment size and the sampling rate. We have thus 
     tc = S / sz  ticks per second to change the rate.
     So, a=(u-v)/tc is the change per 'tick' to go from speed u to
     speed v in 1 second. For other times we multiply with 1/t to
     get a=(u-v)/(t*tc). For example, if we want two seconds, multiplying
     with 1/2 will cut each increment in half and spanning u-v will take
     twice as long. Expand  tc  and we get:

            u - v       sz * (u - v)
     a = ------------ = ------------
         (t * S) / sz      t * S

   */

  a = (sz * (hornRPMFast - hornRPMSlow)) / (hornRevUpSeconds * SampleRateD);

  hornAcceleration =
    (int) (getTableIncrement ((size_t) HFTABLESIZE, a) * 65536.0);

  assert (0.0 < a);

  a = (sz * (hornRPMFast - hornRPMSlow)) / (hornRevDnSeconds * SampleRateD);

  hornDeceleration =
    (int) (getTableIncrement ((size_t) HFTABLESIZE, a) * 65536.0);

  assert (0.0 < a);

#ifdef LPDRUM

  drumSpeedChange = 0;
  drumPosIncSlow = (int)
    (getTableIncrement ((size_t) DFTABLESIZE, drumRPMSlow) * 65536.0);
  drumPosIncFast = (int)
    (getTableIncrement ((size_t) DFTABLESIZE, drumRPMFast) * 65536.0);
  drumPosIncrement = drumPosIncSlow;
  a = (sz * (drumRPMFast - drumRPMSlow)) / (drumRevUpSeconds * SampleRateD);
  drumAcceleration =
    (int) (getTableIncrement ((size_t) DFTABLESIZE, a) * 65536.0);
  a = (sz * (drumRPMFast - drumRPMSlow)) / (drumRevDnSeconds * SampleRateD);
  drumDeceleration =
    (int) (getTableIncrement ((size_t) DFTABLESIZE, a) * 65536.0);

#endif /* LPDRUM */

  /*
   * Here we establish the initial and final positions of certain pointers.
   * dp       Delay write pointer, initialized to the start of the delay line.
   *
   * delayEnd Pointer to an extra sample at the end of the delay line. When
   *          the post-increment address of dp equals delayEnd we know it is
   *          time to wrap back to the start of the delay line.
   *
   * tap1     Pointer to the oldest input value used by the direct filter.
   *          The direct filter models the sound that reaches the listener
   *          first.
   *
   * tap2     Pointer to the oldest input value used by the reflection filter.
   *          The reflection filter models the reflections off two walls,
   *          one placed behind the rotary speaker cabinet and one to the
   *          side. (I tried a third wall, but that diminished rather than
   *          enhanced the effect).
   *
   * tapGuard Points to the first sample in the delay line where the FIR
   *          calculations of tap1 and tap2 will overrun the end of the
   *          delay line. As long as   delay <= tap < tapGuard  the FIR
   *          implementation can use an unchecked calculation. If the
   *          delay line is, for example, 512 samples and the filter has
   *          12 weights, the faster (unchecked) computation is performed
   *          97.6 percent of the time. With two non-overlapping filters
   *          the ratio drops to 95.3 percent which is still quite good,
   *          and can be improved by increasing the length of the delay line.
   */

  dp = delay;
  delayEnd = delay + DLYLNSPLS + 1; /* End of delay line */
#if 0
  tap1 = delay + ((DLYLNSPLS - ( 1 + (HFSPANSIZE - 1))) % DLYLNSPLS);
#else
  /* tap = delay + ((delaylinesz - (FIRlen + offset - 1)) MOD delaylinesz) */
  tap1 = delay + ((DLYLNSPLS - (HFSPANSIZE + 0 - 1)) % DLYLNSPLS);
#endif
#if 0
  tap2 = delay + ((DLYLNSPLS - (29 + (HFSPANSIZE - 1))) % DLYLNSPLS);
#else
  /* In the stereo model the second tap should be identical to tap1,
     unless it is farther away... */
  tap2 = tap1;
#endif
  tapGuard = delay + ((DLYLNSPLS - (HFSPANSIZE - 1)) % DLYLNSPLS);


#ifdef LPDRUM

  drDelayEnd = drDelay + DLYLNSPLS + 1;
  dtap = drDelay;
  dtapGuard = drDelay + ((DLYLNSPLS - (DFSPANSIZE - 1)) % DLYLNSPLS);

/*    fprintf (stderr, "\nDrumINC = %d\n", drumPosIncrement); */

#endif /* LPDRUM */


#ifdef COMMENT
  fprintf (stderr, "\nDrumACC = %d\n", drumAcceleration);
  fprintf (stderr, "\nDrumDEC = %d\n", drumDeceleration);
  fprintf (stderr, "\nHornACC = %d\n", hornAcceleration);
  fprintf (stderr, "HornDEC = %d\n",   hornDeceleration);
  fprintf (stderr, "\ndelay = %x\n", (unsigned int) delay);
  fprintf (stderr, "delayEnd = %x\n", (unsigned int) delayEnd);
  fprintf (stderr, "tapGuard = %x\n", (unsigned int) tapGuard);
#endif /* COMMENT */
}

/*
 *
 */
void setRotaryFast (int isFast) {
  if (isFast != speedIsFast) {
    if (speedIsFast) {
      hornSpeedChange = -1;

#ifdef LPDRUM
      drumSpeedChange = -1;
#endif /* LPDRUM */

    }
    else {
      hornSpeedChange = +1;

#ifdef LPDRUM
      drumSpeedChange = +1;
#endif /* LPDRUM */

    }
    speedIsFast = isFast;
  }
}

/*
 *
 */
void setRotarySusPed (unsigned char b) {
  if (b) {
    setRotaryFast (!speedIsFast);
  }
}

/*
 *
 */
unsigned char * rotsim (unsigned char * inbuf,
			unsigned char * outbuf,
			size_t bufferLengthSamples) {
  int i;
  short * xp = (short *) inbuf;
  short * yp = (short *) outbuf;
  float * tp;
  short * dtp;			/* Drum temporary pointer */

#ifdef CLIP_DETECT
  int over = 0;
#endif /* CLIP_DETECT */

  if (hornSpeedChange) {
    if (hornSpeedChange < 0) {
      hornPosIncrement -= hornDeceleration;
      if (hornPosIncrement <= hornPosIncSlow) {
	hornPosIncrement = hornPosIncSlow;
	hornSpeedChange = 0;
      }
    }
    else {
      hornPosIncrement += hornAcceleration;
      if (hornPosIncFast <= hornPosIncrement) {
	hornPosIncrement = hornPosIncFast;
	hornSpeedChange = 0;
      }
    }
  }

#ifdef LPDRUM

  if (drumSpeedChange) {
    if (drumSpeedChange < 0) {
      drumPosIncrement -= drumDeceleration;
      if (drumPosIncrement <= drumPosIncSlow) {
	drumPosIncrement = drumPosIncSlow;
	drumSpeedChange = 0;
      }
    }
    else {
      drumPosIncrement += drumAcceleration;
      if (drumPosIncFast <= drumPosIncrement) {
	drumPosIncrement = drumPosIncFast;
	drumSpeedChange = 0;
      }
    }
  }

#endif /* LPDRUM */

  for (i = 0; i < bufferLengthSamples; i++) {

#ifdef HFIPOL
    /* Interpolate the two FIRs of the horn filter */
    unsigned int hp = hornPos >> 16;
    unsigned int hq = (hp + 1) & HFTABLESIZEMASK;
    float * wt1v = & (hornFilterTb1[hp][0]);
    float * wt1u = & (hornFilterTb1[hq][0]);
    float * wt2v = & (hornFilterTb2[hp][0]);
    float * wt2u = & (hornFilterTb2[hq][0]);
    float   f    = ((float) (hornPos & 0xFFFF)) / 65536.0;
    float   w1[HFSPANSIZE];
    float   w2[HFSPANSIZE];
    float * wt1;
    float * wt2;
    int j;
#else
    unsigned int hp = hornPos >> 16;
    float * wt1 = & (hornFilterTb1[hp][HFSPANSIZE - 1]);
    float * wt2 = & (hornFilterTb2[hp][HFSPANSIZE - 1]);
#endif /* HFIPOL */

    int Y;
    float Y1;
    float Y2;

#ifdef LPDRUM
    unsigned int drp = drumPos >> 16;
    float * dwt = & (drumFilter[drp][0]);
    float drY;			/* Sum of drum output */

#ifdef DRXOV
    float drX;
    float Z0;
    float Z1;
#endif /* DRXOV */

#endif /* LPDRUM */

#ifdef HFIPOL
    wt1 = &(w1[0]);
    wt2 = &(w2[0]);
    for (j = 0; j < HFSPANSIZE; j++) {
      *wt1++ = (*wt1v) + (f * ((*wt1u) - (*wt1v)));
      *wt2++ = (*wt2v) + (f * ((*wt2u) - (*wt2v)));
      wt1u++;
      wt1v++;
      wt2u++;
      wt2v++;
    }
    wt1--;
    wt2--;
#endif /* HFIPOL */

#ifdef LPDRUM

#ifdef DRXOV

    Z0 = (drxA[0] * (float) (*xp)) + (drxA[1] * drxZ[0]) + (drxA[2] * drxZ[1]);
    Z1 = Z0 + drxZ[0] + drxZ[0] + drxZ[1];
    drxZ[1] = drxZ[0];
    drxZ[0] = Z0;
    Z0 = (drxA[3] * Z1) + (drxA[4] * drxZ[2]) + (drxA[5] * drxZ[3]);
    Z1 = Z0 + drxZ[2] + drxZ[2] + drxZ[3];
    drxZ[3] = drxZ[2];
    drxZ[2] = Z0;
    
#endif /* DRXOV */

    if (dtap < dtapGuard) {

#ifdef DRXOV
      dtap[1] = Z1;
#else
      dtap[1] = (float) (*xp);
#endif /* DRXOV */

      drY = (dwt[1] * dtap[0]) + (dwt[0] * dtap[1]);
      dtap++;			/* No need check wrap here. */

    }
    else {

      drY = (*(dtap++) * dwt[1]);
      if (dtap == drDelayEnd) dtap = drDelay;

#ifdef DRXOV
      dtap[0] = Z1;
#else
      dtap[0] = (float) (*xp);
#endif /* DRXOV */

      drY += (dtap[0] * dwt[0]);

    }

#endif /* LPDRUM */


    *dp++ = (float) (*xp++);		/* Write input sample */


    if (delayEnd <= dp) {	/* Maybe wrap delay line's input pointer */
      dp = delay;
    }

    /* First tap */

    tp = tap1++;		/* Remember current output position */
    if (delayEnd <= tap1) {	/* Maybe wrap the tap1 output pointer */
      tap1 = delay;
    }

    if (tp < tapGuard) {	/* Can we do the fast version? */
      int i;			/* YES */
      Y1 = (*tp++) * (*wt1--);	/* Initialize accumulator */
      for (i = 1; i < HFSPANSIZE; i++) { /* Iterate and accumulate rest */
	Y1 += (*tp++) * (*wt1--);
      }
    }
    else {			/* NO, do slower version */
      int n = delayEnd - tp;	/* Pointer arithmetic */
      int i;
      Y1 = (*tp++) * (*wt1--);	/* Initialize accumulator */
      for (i = 1; i < n; i++) {	/* Iterate and accumulate to end of delay */
	Y1 += (*tp++) * (*wt1--);
      }
      tp = delay;		/* Wrap back to start of delay line */
      n = HFSPANSIZE - i;	/* Calc. count of remaining */
      for (i = 0; i < n; i++) {	/* Iterate and accumulate remaining */
	Y1 += (*tp++) * (*wt1--);
      }
    }

    /* Second tap */

    tp = tap2++;		/* Note position and update tap2. */
    if (delayEnd <= tap2) {	/* Maybe wrap back to start of delay line */
      tap2 = delay;
    }

    if (tp < tapGuard) {	/* Fast version? */
      int i;			/* Yes */
      Y2 = (*tp++) * (*wt2--);
      for (i = 1; i < HFSPANSIZE; i++) {
	Y2 += (*tp++) * (*wt2--);
      }
    }
    else {			/* Slower version */
      int n = delayEnd - tp;
      int i;
      Y2 = (*tp++) * (*wt2--);
      for (i = 1; i < n; i++) {
	Y2 += (*tp++) * (*wt2--);
      }
      tp = delay;
      n = HFSPANSIZE - i;
      for (i = 0; i < n; i++) {
	Y2 += (*tp++) * (*wt2--);
      }
    }


#ifdef CLIP_DETECT
    if (32727.0 < Y1) {
      over++;
    }
    else if (Y1 < -32727.0) {
      over++;
    }
#endif /* CLIP_DETECT */

#ifdef LPDRUM

#ifdef MONO
    *yp++ = (short) (drY + (0.5 * (Y1 + Y2)));
    *yp++ = (short) (drY + (0.5 * (Y1 + Y2)));
#else
    *yp++ = (short) (drY + Y1);
    *yp++ = (short) (drY + Y2);
#endif /* MONO */

#else
				/* Horn only */
#ifdef MONO
    *yp++ = (short) (0.5 * (Y1 + Y2));
    *yp++ = (short) (0.5 * (Y1 + Y2));
#else
    *yp++ = (short) (Y1);
    *yp++ = (short) (Y2);
#endif /* MONO */

#endif /* NOT LPDRUM */

    hornPos = (hornPos + hornPosIncrement) & HFINCTBLMASK;

#ifdef LPDRUM

    drumPos = (drumPos + drumPosIncrement) & DFINCTBLMASK;

#endif /* LPDRUM */
  }

#ifdef CLIP_DETECT
  if (over) {
    putchar ('X');
    fflush (stdout);
  }
#endif /* CLIP_DETECT */

  return outbuf;
}

